% file: routing.t

:- sorts robot;
         x_coordinate;
         y_coordinate;
         direction.

:- variables
         N, N1 :: robot;
         D, D1 ::direction;
         XC, XC1, XC2 :: x_coordinate;
         YC, YC1, YC2 :: y_coordinate;
         X, Y :: computed.
         
:- constants
         0..k :: robot;
         0..maxX :: x_coordinate;
         0..maxY :: y_coordinate;
         left,right,down,up :: direction;
         at_x(robot,x_coordinate),
         at_y(robot,y_coordinate) :: inertialTrueFluent;
         occupied(robot,x_coordinate,y_coordinate) :: defaultFalseFluent;
         move(robot,direction) :: action.

:- macros at(#1,#2,#3) -> (at_x(#1,#2) && at_y(#1,#3)).

% if a robot moves from the grid point (XC,YC) to left then it will be at
% (XC-1, YC). The effects of a robot's moving right, up and down are 
% defined similarly.
move(N,left) causes at_x(N,X) if at_x(N,XC) && X is XC-1 && XC > 0.
move(N,right) causes at_x(N,X) if at_x(N,XC) && X is XC+1 && XC < maxX.
move(N,up) causes at_y(N,Y) if at_y(N,YC) && Y is YC+1 && YC < maxY.
move(N,down) causes at_y(N,Y) if at_y(N,YC) && Y is YC-1 && YC > 0.

% a robot can not move left if it's at the leftmost side of the grid.
% it cannot move right, up and down under similar conditions.
nonexecutable move(N,left) if at_x(N,XC) && XC=<0.    
nonexecutable move(N,right) if at_x(N,XC) && XC>=maxX.
nonexecutable move(N,up) if at_y(N,YC) && YC>=maxY.
nonexecutable move(N,down) if at_y(N,YC) && YC=<0.

% if a robot is at some grid point then it is not anywhere else.
caused -at_x(N,XC)  if at_x(N,XC1) && -(XC=XC1).
caused -at_y(N,YC)  if at_y(N,YC1) && -(YC=YC1).

% a robot cannot move to two different directions at the same.
nonexecutable move(N,D) && move(N,D1) && (D @< D1).

% the points visited by a robot includes its current position and all
% points it had visited by the previous time instant.
caused occupied(N,XC,YC) if at(N,XC,YC).
caused occupied(N,XC,YC) after occupied(N,XC,YC).

% the paths of different robots don't intersect.
never occupied(N,XC,YC) && occupied(N1,XC,YC) && (N < N1).

% robots are prevented from hitting the obstacles.
never at(N,XC,YC) && obstacle(XC1,XC2,YC) && XC =< XC2, XC >= XC1.
never at(N,XC,Y) && obstacle(XC1,XC2,YC) && XC =< XC2, XC >= XC1 && Y is YC+1 && YC < maxY.

% a robot never visits the same point twice.
caused false if at(N,XC,YC) after occupied(N,XC,YC) && -at(N,XC,YC).
